# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from typing import List, Dict
from collections import OrderedDict
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Color import Color
from hanabython.Modules.Colors import Colors
from hanabython.Modules.ConfigurationDeck import ConfigurationDeck
from hanabython.Modules.ConfigurationEmptyClueRule \
    import ConfigurationEmptyClueRule
from hanabython.Modules.ConfigurationHandSize import ConfigurationHandSize
from hanabython.Modules.ConfigurationEndRule import ConfigurationEndRule


class Configuration(Colored):
    """
    A configuration for a game of Hanabi.

    :param deck: the configuration of the deck.
    :param n_clues: the number of clue chips that players have.
    :param n_misfires: the number of misfire chips that players have.
        If :attr:`n_misfires` misfire chips are used, then the game is lost
        immediately (it is not a final warning but really the end of the
        game).
    :param hand_size_rule: the rule used for the initial
        size of the hands.
    :param empty_clue_rule: the rule used about empty clues.
    :param end_rule: the rule used to determine when
        then game is finished.
    :param name: the name of the configuration. Can be None (default value).
        Should not be capitalized (e.g. "my favorite configuration" and not
        "My favorite configuration").

    :var list colors: a list of Color objects. It is the list of keys
        of :attr:`deck`.
    :var int n_colors: the number of colors.
    :var OrderedDict highest: For each color from :attr:`colors`, it gives the
        number on the highest card in that color.
    :var int n_values: the number on the highest card in the whole deck.
    :var list values: the list of possible values (from 1 to :attr:`n_values`).
    :var np.array deck_array: a numpy array of size :attr:`n_colors` *
        :attr:`n_values`. Each row represents the distribution of cards in a
        color. Typically, a row is [3, 2, 2, 2, 1], meaning that there are 3
        ones, 2 twos, etc. Please note that column 0 corresponds to card
        value 1, etc.
    :var int n_cards: the total number of cards in the deck (50 in the
        standard configuration).
    :var int max_score: the maximum possible score (25 in the standard
        configuration).

    >>> cfg = Configuration.W_MULTICOLOR_SHORT
    >>> print(cfg.name)
    with short multicolor (5 cards)
    >>> print(cfg)
    Deck: with short multicolor (5 cards).
    Number of clues: 8.
    Number of misfires: 3.
    Clues rule: empty clues are forbidden.
    End rule: normal.
    >>> print(cfg.hand_size_rule)
    normal
    >>> print(cfg.colors)
    [<Color: B>, <Color: G>, <Color: R>, <Color: W>, <Color: Y>, \
<ColorMulticolor: M>]
    >>> print(cfg.n_colors)
    6
    >>> print(cfg.highest)
    OrderedDict([(<Color: B>, 5), (<Color: G>, 5), (<Color: R>, 5), \
(<Color: W>, 5), (<Color: Y>, 5), (<ColorMulticolor: M>, 5)])
    >>> print(cfg.n_values)
    5
    >>> print(cfg.values)
    [1, 2, 3, 4, 5]
    >>> print(cfg.deck_array)
    [[3 2 2 2 1]
     [3 2 2 2 1]
     [3 2 2 2 1]
     [3 2 2 2 1]
     [3 2 2 2 1]
     [1 1 1 1 1]]
    >>> print(cfg.n_cards)
    55
    >>> print(cfg.max_score)
    30

    Design a configuration manually:

    >>> from hanabython.Modules.ConfigurationDeck import ConfigurationDeck
    >>> from hanabython.Modules.ConfigurationColorContents import (
    ...     ConfigurationColorContents)
    >>> from hanabython.Modules.ConfigurationEmptyClueRule import (
    ...     ConfigurationEmptyClueRule)
    >>> cfg = Configuration(
    ...     deck=ConfigurationDeck(contents=[
    ...         (Colors.BLUE, ConfigurationColorContents([3, 2, 1, 1])),
    ...         (Colors.RED, ConfigurationColorContents([2, 1])),
    ...     ]),
    ...     n_clues=4,
    ...     n_misfires=1,
    ...     hand_size_rule=ConfigurationHandSize.VARIANT_6_3,
    ...     empty_clue_rule=ConfigurationEmptyClueRule.ALLOWED,
    ...     end_rule=ConfigurationEndRule.CROWNING_PIECE
    ... )
    >>> print(cfg)
    Deck: B [3, 2, 1, 1], R [2, 1].
    Number of clues: 4.
    Number of misfires: 1.
    Clues rule: empty clues are allowed.
    End rule: Crowning Piece.
    """
    def __init__(
        self,
        deck: ConfigurationDeck = ConfigurationDeck.NORMAL,
        n_clues: int = 8,
        n_misfires: int = 3,
        hand_size_rule: ConfigurationHandSize = ConfigurationHandSize.NORMAL,
        empty_clue_rule: ConfigurationEmptyClueRule
        = ConfigurationEmptyClueRule.FORBIDDEN,
        end_rule: ConfigurationEndRule = ConfigurationEndRule.NORMAL,
        name: str = None
    ):
        # Parameters
        self.deck = deck
        self.n_clues = n_clues
        self.n_misfires = n_misfires
        self.hand_size_rule = hand_size_rule
        self.empty_clue_rule = empty_clue_rule
        self.end_rule = end_rule
        self.name = name
        # Other attributes
        self.colors = list(deck.keys())                     # type: List[Color]
        self.n_colors = len(self.colors)                    # type: int
        self.highest = OrderedDict([
            (c, len(deck[c])) for c in self.colors
        ])                                      # type: OrderedDict[Color, int]
        self.n_values = max(self.highest.values())          # type: int
        self.values = list(range(1, self.n_values + 1))     # type: List[int]
        self.deck_array = np.array([
            deck[c] + [0] * (self.n_values - len(deck[c]))
            for c in self.colors
        ])                                                  # type: np.array
        self.n_cards = np.sum(self.deck_array)              # type: int
        self.max_score = sum(self.highest.values())         # type: int
        # Conversion
        self._i_from_c_name = {
            c.name: i for i, c in enumerate(self.colors)
        }                                               # type: Dict[Color, int]

    def __repr__(self) -> str:
        if self.name:
            return '<Configuration: %s>' % self.name
        return (
            '<Configuration: %r, n_clues=%s, n_misfires=%s, '
            '%r, %r, %r>'
            % (self.deck, self.n_clues, self.n_misfires,
               self.hand_size_rule, self.empty_clue_rule, self.end_rule)
        )

    def colored(self) -> str:
        return '\n'.join([
            'Deck: %s.' % self.deck.colored(),
            'Number of clues: %s.' % self.n_clues,
            'Number of misfires: %s.' % self.n_misfires,
            'Clues rule: %s.' % self.empty_clue_rule.colored(),
            'End rule: %s.' % self.end_rule.colored()
        ])

    def i_from_c(self, c: Color) -> int:
        """
        Finds index from a color (for example in :attr:`deck_array`).

        :param c: a color.

        :return: the corresponding index.

        >>> Configuration.STANDARD.i_from_c(Colors.BLUE)
        0
        """
        return self._i_from_c_name[c.name]

    # noinspection PyMethodMayBeStatic
    def i_from_v(self, v: int) -> int:
        """
        Finds index from a value (for example in :attr:`deck_array`).

        :param v: the value (typically 1 to 5).

        :return: the corresponding index (typically 0 to 4).

        >>> Configuration.STANDARD.i_from_v(1)
        0
        """
        return v - 1

    #:
    STANDARD = None
    #:
    W_SIXTH = None
    #:
    W_SIXTH_SHORT = None
    #:
    W_MULTICOLOR = None
    #:
    W_MULTICOLOR_SHORT = None
    #:
    EIGHT_COLORS = None


Configuration.STANDARD = Configuration(name='standard')
Configuration.W_SIXTH = Configuration(
    deck=ConfigurationDeck.W_SIXTH,
    name=ConfigurationDeck.W_SIXTH.name)
Configuration.W_SIXTH_SHORT = Configuration(
    deck=ConfigurationDeck.W_SIXTH_SHORT,
    name=ConfigurationDeck.W_SIXTH_SHORT.name)
Configuration.W_MULTICOLOR = Configuration(
    deck=ConfigurationDeck.W_MULTICOLOR,
    name=ConfigurationDeck.W_MULTICOLOR.name)
Configuration.W_MULTICOLOR_SHORT = Configuration(
    deck=ConfigurationDeck.W_MULTICOLOR_SHORT,
    name=ConfigurationDeck.W_MULTICOLOR_SHORT.name)
Configuration.EIGHT_COLORS = Configuration(
    deck=ConfigurationDeck.EIGHT_COLORS,
    name=ConfigurationDeck.EIGHT_COLORS.name)


if __name__ == '__main__':
    my_cfg = Configuration.EIGHT_COLORS
    my_cfg.test_str()

    print('\nAttributes: ')
    for k in my_cfg.__dict__.keys():
        if not k.startswith('_'):
            print(k + ': ', my_cfg.__dict__[k])

    print('\nOther standard configurations: ')
    print(Configuration.W_SIXTH.colored() + '\n')
    print(Configuration.W_SIXTH_SHORT.colored() + '\n')
    print(Configuration.W_MULTICOLOR.colored() + '\n')
    print(Configuration.W_MULTICOLOR_SHORT.colored() + '\n')

    print('\nA manual configuration: ')
    from hanabython.Modules.ConfigurationColorContents \
        import ConfigurationColorContents
    my_cfg = Configuration(
        deck=ConfigurationDeck(contents=[
            (Colors.BLUE, ConfigurationColorContents([3, 2, 1, 1])),
            (Colors.RED, ConfigurationColorContents([2, 1])),
        ]),
        n_clues=4,
        n_misfires=1,
        hand_size_rule=ConfigurationHandSize.VARIANT_6_3,
        empty_clue_rule=ConfigurationEmptyClueRule.ALLOWED,
        end_rule=ConfigurationEndRule.CROWNING_PIECE
    )
    print(my_cfg.colored())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from hanabython.Modules.Colored import Colored
from hanabython.Modules.StringUtils import uncolor
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.Color import Color
from hanabython.Modules.Card import Card


class Board(Colored):
    """
    The board (cards successfully played) in a game of Hanabi.

    :param cfg: the configuration of the game.

    :var np.array altitude: indicates the highest card played in each color.
        E.g. with color ``c`` of index ``i``, ``altitude[i]`` is the value
        of the highest card played in color ``c``. The correspondence between
        colors and indexes is the one provided by :attr:`cfg`.

    >>> from hanabython.Modules.Configuration import Configuration
    >>> board = Board(Configuration.STANDARD)
    >>> print(board.altitude)
    [0 0 0 0 0]
    """

    def __init__(self, cfg: Configuration):
        self.cfg = cfg
        self.altitude = np.zeros(self.cfg.n_colors, dtype=int)  # type: np.array

    def __repr__(self) -> str:
        return '<Board: %s>' % self.str_compact()

    def colored(self) -> str:
        return self.colored_fixed_space()

    def str_compact(self) -> str:
        """
        Convert to string in "compact" layout.

        :return: a representation of the board.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> board = Board(Configuration.STANDARD)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board.str_compact())
        G1 G2 Y1 Y2 Y3 Y4 Y5
        """
        return uncolor(self.colored_compact())

    def colored_compact(self) -> str:
        """
        Colored version of :meth:`str_compact`.
        """
        if self.score == 0:
            return 'No card on the board yet'
        return ' '.join([
            c.color_str(self._str_one_color(i, c))
            for i, c in enumerate(self.cfg.colors)
            if self.altitude[i] > 0
        ])

    def str_fixed_space(self) -> str:
        """
        Convert to string in "fixed-space" layout.

        :return: a representation of the board.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> board = Board(Configuration.STANDARD)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board.str_fixed_space())
        B -         G 1 2       R -         W -         Y 1 2 3 4 5
        """
        return uncolor(self.colored_fixed_space())

    def colored_fixed_space(self) -> str:
        """
        Colored version of :meth:`str_fixed_space`.
        """
        length = 1 + 2 * self.cfg.n_values
        return ' '.join([
            c.color_str(self._str_one_color_factorized(i, c).ljust(length))
            for i, c in enumerate(self.cfg.colors)
        ])

    def str_multi_line(self) -> str:
        """
        Convert to string in "multi-line" layout.

        :return: a representation of the board.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> board = Board(Configuration.STANDARD)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board.str_multi_line())
        -
        G1 G2
        -
        -
        Y1 Y2 Y3 Y4 Y5
        """
        return uncolor(self.colored_multi_line())

    def colored_multi_line(self) -> str:
        """
        Colored version of :meth:`str_multi_line`.
        """
        return '\n'.join([
            c.color_str(self._str_one_color(i, c))
            for i, c in enumerate(self.cfg.colors)
        ])

    def str_multi_line_compact(self) -> str:
        """
        Convert to string in "compact multi-line" layout.

        :return: a representation of the board.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> board = Board(Configuration.STANDARD)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board.str_multi_line_compact())
        G1 G2
        Y1 Y2 Y3 Y4 Y5
        """
        return uncolor(self.colored_multi_line_compact())

    def colored_multi_line_compact(self) -> str:
        """
        Colored version of :meth:`str_multi_line_compact`.
        """
        if self.score == 0:
            return 'No card on the board yet'
        return '\n'.join([
            c.color_str(self._str_one_color(i, c))
            for i, c in enumerate(self.cfg.colors)
            if self.altitude[i] > 0
        ])

    # noinspection PyProtectedMember
    def _str_one_color(self, i: int, c: Color) -> str:
        """
        Convert one color to string.

        :param i: index of the color.
        :param c: the color.

        :return: a representation of the cards played in this color.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> cfg = Configuration.STANDARD
        >>> board = Board(cfg)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board._str_one_color(i=4, c=cfg.colors[4]))
        Y1 Y2 Y3 Y4 Y5
        """
        if self.altitude[i] == 0:
            return '-'
        return ' '.join([
            str(Card(c, j)) for j in range(1, self.altitude[i] + 1)
        ])

    # noinspection PyProtectedMember
    def _str_one_color_factorized(self, i: int, c: Color) -> str:
        """
        Same as :meth:`_str_one_color`, but with the color symbol only once.

        :param i: index of the color.
        :param c: the color.

        :return: a representation of the cards played in this color.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> cfg = Configuration.STANDARD
        >>> board = Board(cfg)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board._str_one_color_factorized(i=4, c=cfg.colors[4]))
        Y 1 2 3 4 5
        """
        if self.altitude[i] == 0:
            return c.symbol + ' -'
        return c.symbol + ' ' + ' '.join([
            str(j) for j in range(1, self.altitude[i] + 1)
        ])

    def try_to_play(self, card: Card) -> bool:
        """
        Try to play a card on the board.

        :param card: the card.

        :return: True if the card is successfully played on the board, False
            otherwise (i.e. if it leads to a misfire).

        >>> from hanabython.Modules.Configuration import Configuration
        >>> from hanabython.Modules.Card import Card
        >>> board = Board(Configuration.STANDARD)
        >>> for s in ['B1', 'B2', 'Y1', 'Y3', 'B1']:
        ...     board.try_to_play(Card(s))
        True
        True
        True
        False
        False
        >>> print(board.str_compact())
        B1 B2 Y1
        """
        i_c = self.cfg.i_from_c(card.c)
        if card.v == self.altitude[i_c] + 1:
            self.altitude[i_c] += 1
            return True
        else:
            return False

    @property
    def score(self) -> int:
        """
        The current score.

        :return: the sum of the altitudes reached in all colors.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> cfg = Configuration.STANDARD
        >>> board = Board(cfg)
        >>> for s in ['G1', 'G2', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5']:
        ...     _ = board.try_to_play(Card(s))
        >>> print(board.score)
        7
        """
        return int(np.sum(self.altitude))


if __name__ == '__main__':
    my_board = Board(Configuration.W_MULTICOLOR_SHORT)
    for s in ['B1', 'B2', 'M1', 'M3', 'B1']:
        print('Try to play %s: ' % s, my_board.try_to_play(Card(s)))
    print()
    my_board.test_str()

    print('\nAll layout styles (colored version):')
    print('Compact: ')
    print(my_board.colored_compact())
    print('\nFixed space: ')
    print(my_board.colored_fixed_space())
    print('\nMulti-line compact: ')
    print(my_board.colored_multi_line_compact())
    print('\nMulti-line: ')
    print(my_board.colored_multi_line())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Colored import Colored


class Action(Colored):
    """
    An action performed by a player (Throw, Play a card, Clue or Forfeit).

    In the end-user interfaces (including methods ``colored``), "throw" should
    be called "discard" and "play a card" can be called "play" (to be consistent
    with the official rules). In the code however, we prefer to use "throw"
    (to distinguish from other forms of discards, for example after a misfire)
    and "play a card" (to distinguish from simply playing in general).

    :param category: can be :attr:`Action.THROW`, :attr:`Action.PLAY_CARD`,
        :attr:`Action.CLUE` or :attr:`Action.FORFEIT`.

    Generally, only subclasses are instantiated. Cf. :class:`ActionThrow`,
    :class:`ActionPlayCard`, :class:`ActionClue` and :class:`ActionForfeit`.
    """

    #:
    THROW = 0
    #:
    PLAY_CARD = 1
    #:
    CLUE = 2
    #:
    FORFEIT = 3
    #: Possibles categories of action.
    CATEGORIES = {THROW, PLAY_CARD, CLUE, FORFEIT}

    def __init__(self, category: int):
        if category not in Action.CATEGORIES:
            raise ValueError('Unknown action category: ', category)
        self.category = category



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Union
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Color import Color
from hanabython.Modules.Colors import Colors


class Clue(Colored):
    """
    A clue.

    :param x: the clue (value or color).

    :var int category: can be either Clue.VALUE or Clue.COLOR.

    >>> clue = Clue(1)
    >>> print(clue)
    1
    >>> clue.category == Clue.VALUE
    True
    >>> clue = Clue(Colors.RED)
    >>> print(clue)
    R
    >>> clue.category == Clue.COLOR
    True
    """

    #: Category for a clue by value.
    VALUE = 0
    #: Category for a clue by color.
    COLOR = 1

    def __init__(self, x: Union[int, Color]):
        self.x = x
        self.category = Clue.VALUE if type(x) == int else Clue.COLOR

    def colored(self) -> str:
        if self.category == self.VALUE:
            return str(self.x)
        else:
            return self.x.colored()


if __name__ == '__main__':
    my_clue = Clue(1)
    my_clue.test_str()
    print()
    my_clue = Clue(Colors.BLUE)
    my_clue.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from hanabython.Modules.PlayerBase import PlayerBase
from hanabython.Modules.Action import Action
from hanabython.Modules.ActionClue import ActionClue
from hanabython.Modules.ActionPlayCard import ActionPlayCard
from hanabython.Modules.ActionThrow import ActionThrow
from hanabython.Modules.ActionForfeit import ActionForfeit
from hanabython.Modules.Clue import Clue
from IPython.display import clear_output
from time import sleep


class PlayerHumanText(PlayerBase):
    """
    User interface for a human player in text mode (terminal or notebook).

    :param ipython: use `True` when using the player in a notebook. This
        fixes a problem between ``clear_output`` and ``input``.

    >>> antoine = PlayerHumanText('Antoine', ipython=True)
    """

    def __init__(self, name: str, ipython=False):
        super().__init__(name)
        self.ipython = ipython

    # *** General methods about actions ***

    def receive_turn_begin(self) -> None:
        """
        We pause, then we inform the player of the most recent events.
        """
        print('\n' * 40)
        if self.ipython:
            clear_output()
            sleep(0.5)  # Essential line to prevent strange behavior in Jupyter!
        input('%s is going to play (hit Enter).\n' % self.name)
        print(self.colored())

    def choose_action(self) -> Action:
        """
        The human player gets to choose an action.
        """
        category = None             # type: int
        i = None                    # type: int
        cat_dico = {'C': Action.CLUE, 'P': Action.PLAY_CARD,
                    'D': Action.THROW, 'F': Action.FORFEIT}
        while True:
            if category is None:
                cat_str = input('\nWhat action? (C = Clue, P = Play, '
                                'D = Discard, F = Forfeit)\n')
                if not cat_str:  # Nothing in the input => try again
                    continue
                cat_str = cat_str[0].capitalize()
                if cat_str in cat_dico.keys():
                    category = cat_dico[cat_str]
            elif category in {Action.PLAY_CARD, Action.THROW}:
                k_str = input('What card? (1 = leftmost, etc.)\n')
                try:
                    k = int(k_str) - 1
                except ValueError:  # Not valid => cancel
                    category = None
                    continue
                if category == Action.PLAY_CARD:
                    return ActionPlayCard(k)
                else:
                    return ActionThrow(k)
            elif category == Action.FORFEIT:
                confirm_str = input('Do you confirm forfeit? (Y/N)\n')
                if not confirm_str:  # Nothing in the input => cancel
                    category = None
                    continue
                if confirm_str[0].capitalize() == 'Y':
                    return ActionForfeit()
                else:
                    category = None  # Not confirmed ==> cancel
            elif category == Action.CLUE:
                if i is None:
                    if self.n_players == 2:
                        i = 1
                        continue
                    i_str = input('What player? (1 = next player, etc.)\n')
                    try:
                        i = int(i_str)
                    except ValueError:  # Not valid => cancel
                        category = None
                else:
                    clue_str = input('What clue? (B, G, ..., 1, 2, ...)\n')
                    try:
                        return ActionClue(i, Clue(int(clue_str)))
                    except ValueError:
                        pass
                    try:
                        clue_str = clue_str[0].capitalize()
                        clue = Clue([
                            c for c in self.cfg.colors if c.symbol == clue_str
                        ][0])
                        return ActionClue(i, clue)
                    except IndexError:
                        category = None
                        i = None
            else:
                logging.critical('This case should never happen.')

    def receive_action_legal(self) -> None:
        """
        We forget the previous events.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> antoine = PlayerHumanText('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.log('Donald does something.')
        >>> antoine.recent_events
        'Donald does something.'
        >>> # Here, Antoine would choose his own action. Then...
        >>> antoine.receive_action_legal()
        >>> antoine.log("Antoine's action has such and such consequences.")
        >>> antoine.recent_events
        "Antoine's action has such and such consequences."
        """
        self.log_forget()

    def receive_action_illegal(self, s: str) -> None:
        print(s)

    def receive_turn_finished(self) -> None:
        """
        We inform the player of the most recent events, i.e. the consequences
        of her actions. Then we pause (unless this string was empty).
        Finally, we forget these recent events.
        """
        print(self.recent_events)
        if self.recent_events:
            input("Your turn is over (hit Enter).\n")
        self.log_forget()

    # *** End of game ***

    def receive_lose(self, score: int) -> None:
        """
        We print and forget the recent (unfortunate) events.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> antoine = PlayerHumanText('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_lose(score=0)
        Antoine's team loses.
        Score: 0.
        <BLANKLINE>
        >>> antoine.recent_events
        ''
        """
        super().receive_lose(score)
        print(self.recent_events)
        self.log_forget()

    def receive_game_exhausted(self, score: int) -> None:
        """
        We print and forget the recent events.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> antoine = PlayerHumanText('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_game_exhausted(score=23)
        Antoine's team has reached the end of the game.
        Score: 23.
        <BLANKLINE>
        >>> antoine.recent_events
        ''
        """
        super().receive_game_exhausted(score)
        print(self.recent_events)
        self.log_forget()

    def receive_win(self, score: int) -> None:
        """
        We print and forget the recent (cheerful) events.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> antoine = PlayerHumanText('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_win(score=25)
        Antoine's team wins!
        Score: 25.
        <BLANKLINE>
        >>> antoine.recent_events
        ''
        """
        super().receive_win(score)
        print(self.recent_events)
        self.log_forget()


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Callable
from hanabython.Modules.Colored import Colored


class ConfigurationHandSize(Colored):
    """
    A rule for the initial size of the players' hands.

    :param f: a callable that, to a number of players, associates a number of
        cards.
    :param name: the name of the configuration. Can be None (default value).
        Should not be capitalized (e.g. "my favorite configuration" and not
        "My favorite configuration").

    >>> cfg = ConfigurationHandSize.NORMAL
    >>> print(cfg)
    normal
    >>> cfg = ConfigurationHandSize(f=lambda n: 9 - n)
    >>> print(cfg)
    7 for 2p, 6 for 3p, 5 for 4p, 4 for 5p
    """

    def __init__(self, f: Callable[[int], int], name: str = None):
        self.f = f
        self.name = name

    def colored(self) -> str:
        if self.name is None:
            return ', '.join([
                '%s for %sp' % (self.f(n), n) for n in range(2, 6)
            ])
        else:
            return self.name

    #: Normal rule for hand size (5 for 3- players, 4 for 4+ players).
    NORMAL = None
    #: Experimental variant for hand size (6 for 2 players, 3 for 5+ players).
    VARIANT_6_3 = None


ConfigurationHandSize.NORMAL = ConfigurationHandSize(
    f=lambda n: 5 if n <= 3 else 4,
    name='normal'
)
ConfigurationHandSize.VARIANT_6_3 = ConfigurationHandSize(
    f=lambda n: 3 if n >= 5 else 8 - n,
    name='experimental (6 for 2 players, 3 for 5 players)'
)


if __name__ == '__main__':
    print('A configuration with a name:')
    my_cfg = ConfigurationHandSize.NORMAL
    my_cfg.test_str()

    print('\nA configuration with no name:')
    my_cfg = ConfigurationHandSize(f=lambda n: 9 - n)
    my_cfg.test_str()

    import doctest
    doctest.testmod()
